from .combo import combo
from .. import elements as el
from ..elements.style_utils import to_rgba, rgba_to_hex


def line_chart(*data, x_axis="category", **props):
    return basic_chart("line", *data, x_axis=x_axis, **props)


def bar_chart(*data, x_axis="category", **props):
    return basic_chart("bar", *data, x_axis=x_axis, **props)


def scatter_chart(*data, **props):
    return basic_chart("scatter", *data, **props)


def parse_axis(axis):
    """
    "value", "time", "log", "category", or a list of named categories
    """
    if axis is None:
        return dict(show=False)
    elif axis in ("value", "time", "log", "category"):
        return dict(type=axis)
    elif type(axis) in (list, tuple):
        return dict(type="category", data=axis)

    raise ValueError("Invalid chart axis.")


def parse_tooltip(tooltip):
    tooltip_options = None
    if tooltip is not None:
        tooltip_options = dict()
        if tooltip in ("axis", "item"):
            tooltip_options["trigger"] = tooltip
        else:
            raise ValueError("Invalid tooltip.")
    return tooltip_options


def parse_series(chart_type, data, series_names):
    return [
        dict(
            type=chart_type,
            data=series,
            name=index if series_names is None else series_names[index],
        )
        for index, series in enumerate(data)
    ]


def complementary_color(color):
    r, g, b, a = to_rgba(color)
    rgba = (255 - r, 255 - g, 255 - b, a)
    return rgba_to_hex(rgba)


class basic_chart(combo):
    """
    Creates a simple echart based on one or more series,
    all of the same type (e.g. line, chart, scatter).
    Use x_axis=dict(data=[...]) to create an axis
    that has categories. Otherwise data values will be used.
    """

    def __init__(
        self,
        chart_type,  # scatter, line, bar
        *data,  # Each separate series is a list/tuple
        series_names=None,  # Used for tooltips and legends.
        x_axis="value",  # See parse_axis for valid format
        y_axis="value",
        rotate=False,
        tooltip=None,  # None, "axis", or "item" (where it's triggered)
        color=None,
        **props,  # width, height
    ):
        if len(data) == 1 and isinstance(data[0], dict):
            x_axis = list(data[0].keys())
            data = (list(data[0].values()),)

        state = el.state(
            x_axis=x_axis,
            y_axis=y_axis,
            rotate=rotate,
            data=data,
            series_names=series_names,
            tooltip=tooltip,
            chart_type=chart_type,
            color=color,
        )

        if len(state.data) == 0:
            raise ValueError("Chart must have data.")

        if state.chart_type not in ("line", "bar", "scatter"):
            raise ValueError("Invalid chart type.")

        x_axis = parse_axis(state.x_axis)
        y_axis = parse_axis(state.y_axis)

        if (
            type(data[0][0]) not in (list, tuple)
            and x_axis["type"] != "category"
            and y_axis["type"] != "category"
        ):
            raise ValueError(
                "Invalid data. Use tuple values or make axis type `category`."
            )

        tooltip_options = parse_tooltip(state.tooltip)
        series = parse_series(chart_type, state.data, state.series_names)

        if state.rotate:
            x_axis, y_axis = y_axis, x_axis

        theme = el.theme()

        chart = el.echart(
            options=dict(
                xAxis=x_axis,
                yAxis=y_axis,
                series=series,
                tooltip=tooltip_options,
                #
                # grid is used to remove spacing.
                grid=dict(top=8, right=8, bottom=0, left=0, containLabel=True),
                #
                # Use hyperdiv colors by default.
                color=state.color or dict(auto=True, seeds=[theme.primary_color]),
            ),
        )

        super().__init__(
            (state, ["x_axis", "y_axis", "series_names", "data", "tooltip", "rotate"]),
            (
                chart,
                [
                    "data_clicked",
                    "width",
                    "height",
                    "min_width",
                    "min_height",
                    "grow",
                    "shrink",
                    "border",
                    "border_radius",
                    "padding",
                    "background_color",
                    "hover_background_color",
                    "active_background_color",
                    "font_color",
                    "font_weight",
                ]
                + el.input_props,
            ),
            init=props,
        )
