from .combo import combo
from .. import elements as el

props_to_forward = (
    "value",
    "valid",
    "required",
    "background_color",
    "border",
    "outline",
    "padding",
    "grow",
)


class radio_group(combo):
    def __init__(
        self,
        label,
        value,
        values,
        disabled=False,
        required=False,
        valid=True,
        **props,
    ):
        self._state = el.state(
            value=value,
            disabled=disabled,
            required=required,
            valid=valid,
            edited=False,
        )

        radio_color = el.theme().primary_color
        font_color = el.theme().font_color
        background_color = None
        if not valid and self._state.edited:
            font_color = el.theme().error_color
            radio_color = el.theme().error_color
            background_color = el.lighten(el.theme().error_color, 130)

        container = el.container(
            role="radiogroup",
            font_color=font_color,
            background_color=background_color,
        )

        super().__init__(
            (
                container,
                [
                    "gap",
                    "direction",
                    "grow",
                    "border",
                    "outline",
                    "padding",
                    "background_color",
                ]
                + el.input_props,
            ),
            self._state,
            init=props,
        )

        if len(set(values)) != len(values):
            raise ValueError("Duplicate values are not allowed in radio groups.")

        with container:
            if len(label) > 0:
                with el.container(direction="horizontal", align="end", gap=2):
                    el.text(label, font_size="small")
                    if self._state.required:
                        el.text("*", font_color=el.theme().error_color)
            for v in values:
                with el.scope(v):
                    # self.value is accessible because self._state is
                    # in combo.__init__()
                    selected = v == self.value
                    item_disabled = self._item_disabled(v)

                    t = el.radio(
                        group=container.key,
                        label=v,
                        selected=selected,
                        disabled=item_disabled,
                        radio_color=radio_color,
                    )

                    # If the radio is reset, it will
                    # also call reset on this group
                    t.resettable_parent = self

                    # Here we check which item is selected.
                    # Note: We use the "selected" prop
                    # rather than using "clicked"
                    # because radios can be selected
                    # without clicks (tabs + enter).
                    if t.selected:
                        self.value = v

                    if t.edited:
                        self.edited = True

    def _item_disabled(self, item_value):
        if type(self.disabled) in [list, tuple]:
            return item_value in self.disabled
        elif type(self.disabled) is str:
            return item_value == self.disabled

        return self.disabled

    @property
    def is_empty(self):
        return False

    def reset(self):
        self._state.reset_prop("value")
        self._state.reset_prop("edited")

    @property
    def selected(self):
        return self.value

    @selected.setter
    def selected(self, value):
        self.value = value
        self.edited = True
