def is_dom(node):
    return isinstance(node, dict) and node.get("_dom") is True


def is_dom_list(node):
    return isinstance(node, list) and len(node) > 0 and is_dom(node[0])


def diff_dicts(src, dest):
    """
    Assumes src and dest are not-None
    - deep diff dom if available on both sides.
    - otherwise remove/update fields
    - if src or dest is None, a straight update
    """
    diff_chunks = list()

    keys = sorted({**src, **dest}.keys())

    for k in keys:
        if k not in src:
            diff_chunks.append(("update_key", k, dest[k]))
            continue
        elif k not in dest:
            diff_chunks.append(("remove_key", k))
            continue

        sk = src[k]
        dk = dest[k]

        if is_dom_list(sk) and is_dom_list(dk):
            subdiff = diff_lists(sk, dk)
            if len(subdiff) > 0:
                diff_chunks.append(("subdiff_list", k, subdiff))
        # elif is_dom(sk) and is_dom(dk) and src["key"] == dest["key"]:
        #     subdiff = diff_dicts(sk, dk)
        #     if len(subdiff) > 0:
        #         diff_chunks.append(("subdiff_dict", k, subdiff))
        elif sk != dk:
            diff_chunks.append(("update_key", k, dk))

    return diff_chunks


def diff_lists(src, dest):
    """Emits a list of commands (diff) that when applied to `src`, `src`
    will become equal to `dest`. The diff can be applied using
    `patch`.

    Assumptions:

    - Every node has a unique "key". If a node in src has the same id
      as a node in dest, we can assume they are the same node.

    - If a node has children fields (fields recursively holding dom)
      the node has a `children` attribute which is a list of field
      names that hold children.

    The emitted diff is actually a tree. One of the diff commands is
    "subdiff" which contains a diff to be applied recursively to a
    node.
    """
    diff_chunks = list()
    src_idx = 0
    dest_idx = 0
    while True:
        try:
            src_elem = src[src_idx]
        except IndexError:
            # Reached the end of the src dom. To get the dest dom from
            # the src dom, we need to append the rest of the dest dom
            # to the end of the src dom.
            #
            if dest_idx < len(dest):
                diff_chunks.append(("insert", dest_idx, dest[dest_idx:]))
            break

        try:
            dest_elem = dest[dest_idx]
        except IndexError:
            # Reached the end of the dest dom. To get the src dom from
            # it, we need to truncate the rest of the src dom.
            diff_chunks.append(("delete", dest_idx, len(src) - src_idx))
            break

        # Found two corresponding nodes. We emit diff chunks to update
        # the keys in src to match those in dest. If one of the keys
        # is a "children" key, we diff recursively.
        if src_elem["key"] == dest_elem["key"]:
            dict_diff = diff_dicts(src_elem, dest_elem)
            if len(dict_diff) > 0:
                diff_chunks.append(("dict_diff", dest_idx, dict_diff))
            src_idx += 1
            dest_idx += 1
        else:
            # Three possible situations starting at indices src_idx
            # and dest_idx:
            #   (a)
            #     src:   x ...
            #     dest:  y ... x ...
            #     If x occurs later in dest, (y ...) needs to be
            #     inserted before x in src.
            #   (b)
            #     src:   y ... x ...
            #     dest:  x ...
            #     If x occurs later in src, Here (y ...) needs to be
            #     removed from src.
            #   (c)
            #     src:   x ...
            #     dest:  y ...
            #     Where y doesn't occur in src and x doesn't occur in
            #     dest. In this case we replace the rest of src with
            #     the rest of dest.
            si = src_idx
            di = dest_idx
            found = False

            # We scan for situations (a) and (b) at the same time by
            # advancing the src and dest indices in lock step.
            while di < len(dest) or si < len(src):
                if di < len(dest):
                    # (a)
                    if dest[di]["key"] == src_elem["key"]:
                        diff_chunks.append(("insert", dest_idx, dest[dest_idx:di]))
                        dest_idx = di
                        found = True
                        break
                if si < len(src):
                    # (b)
                    if src[si]["key"] == dest_elem["key"]:
                        diff_chunks.append(("delete", dest_idx, si - src_idx))
                        src_idx = si
                        found = True
                        break
                si += 1
                di += 1

            if found:
                continue

            # (c)
            diff_chunks.append(("delete", dest_idx, len(src) - src_idx))
            diff_chunks.append(("insert", dest_idx, dest[dest_idx:]))
            break

    return diff_chunks


def patch_list(base, diff):
    for d in diff:
        if d[0] == "insert":
            _, idx, dom = d
            base[idx:idx] = dom
        elif d[0] == "delete":
            _, idx, num_items = d
            base[idx : idx + num_items] = []
        elif d[0] == "dict_diff":
            _, idx, subdiff = d
            patch_dict(base[idx], subdiff)
        else:
            raise Exception("unknown type of diff chunk")


def patch_dict(base, diff):
    for d in diff:
        if d[0] == "update_key":
            _, k, new_value = d
            base[k] = new_value
        elif d[0] == "remove_key":
            _, k = d
            base.pop(k, None)
        elif d[0] == "subdiff_list":
            _, k, subdiff = d
            patch_list(base[k], subdiff)
        # elif d[0] == "subdiff_dict":
        #     _, k, subdiff = d
        #     patch_dict(base[k], subdiff)
        else:
            raise Exception(f"Unknown diff chunk type {d[0]}")
