import yappi
import traceback
import uuid
import os
import threading
import json
from tornado.websocket import WebSocketHandler
from tornado.web import Application, StaticFileHandler
from tornado.ioloop import IOLoop
import tornado.log
from .context import Context, logger
from .utils import static_directory, strip_path_suffix

tornado.log.enable_pretty_logging()
ioloop = IOLoop.current()


class SavedContexts:
    def __init__(self):
        self.contexts = dict()
        self.lock = threading.RLock()

    def add_context(self, client_id, context):
        with self.lock:
            self.contexts[client_id] = context

    def get_context(self, client_id):
        with self.lock:
            return self.contexts.get(client_id)


class Clients:
    def __init__(self):
        self.clients = dict()
        self.lock = threading.RLock()

    def add_client(self, client):
        with self.lock:
            self.clients[client.client_id] = client

    def remove_client(self, client_id):
        with self.lock:
            self.clients.pop(client_id)

    def get_clients(self):
        with self.lock:
            return list(self.clients.values())


def shutdown():
    clients = ClientConnection.clients.get_clients()
    logger.info(f"Shutting down {len(clients)} connections.")
    for client in ClientConnection.clients.get_clients():
        client.context.stop()


class ClientConnection(WebSocketHandler):
    clients = Clients()
    saved_contexts = SavedContexts()

    """A new client connection object is created every time a client
    connects.
    """

    def __init__(self, application, request):
        super().__init__(application, request)
        self.opened = threading.Event()
        self.context = None

        client_id = self.get_argument("clientId", None)
        context = None

        if client_id:
            # Reconnecting; see if the clientId is valid and
            # corresponds to a saved context.
            try:
                client_id = uuid.UUID(client_id)
                context = ClientConnection.saved_contexts.get_context(client_id)
            except Exception:
                pass

        if context:
            self.client_id = client_id
            self.context = context
            logger.info(f"{self.client_id} - Reconnecting to existing context.")
        else:
            self.client_id = uuid.uuid4()
            logger.info(f"{self.client_id} - Fresh connection.")

            self.context = self.create_context()
            if not self.context:
                return
            ClientConnection.saved_contexts.add_context(self.client_id, self.context)

        updates = self.get_argument("updates", None)
        if updates:
            try:
                self.apply_updates(json.loads(updates))
            except Exception:
                logger.warn(f"Malformed updates argument: {updates}")

        ClientConnection.clients.add_client(self)
        self.context.start(self)

    def apply_updates(self, updates):
        self.context.set_state_bulk(
            [
                (update["key"], update["propName"], update["propValue"])
                for update in updates
            ]
        )

    def create_context(self):
        return Context(main_function)

    def open(self):
        self.opened.set()

    def on_message(self, message_batch_json):
        """Handles updates coming from the client."""
        if not self.context:
            return
        try:
            message_batch = json.loads(message_batch_json)
            for msg in message_batch:
                if msg["type"] == "update":
                    self.apply_updates(msg["updates"])

        except Exception:
            logger.exception("Failure on recv.")

    def on_close(self):
        logger.info(f"{self.client_id} - Connection closed")
        # TODO: Don't join on the ioloop.
        if not self.context:
            return
        self.context.stop()
        self.clients.remove_client(self.client_id)

    def send(self, message):
        def _write():
            try:
                self.write_message(json.dumps(message))
            except Exception:
                logger.exception("Write to client failed")

        ioloop.add_callback(_write)

    def get_compression_options(self):
        # See https://www.tornadoweb.org/en/stable/websocket.html
        # Enables compression:
        return dict()


def force_global_refresh():
    clients = ClientConnection.clients.get_clients()
    for client in clients:
        client.context.refresh()


class HyperdivStaticFileHandler(StaticFileHandler):
    def get_absolute_path(self, root, path):
        try:
            os.stat(os.path.join(root, path))
        except FileNotFoundError:
            path = "index.html"
        return super().get_absolute_path(root, path)


main_function = None


def get_public_folder_path():
    path = os.path.join(
        strip_path_suffix(__file__, 2),
        "public",
    )

    # We're in the public distribution
    if os.path.exists(path):
        return path

    # We're in development
    return os.path.join(
        strip_path_suffix(__file__, 3),
        "frontend",
        "public",
    )


def run(main_fun, http_routes=None):
    """The main entrypoint. Register the given main function and starts the asdf
    websocket server.
    """
    global main_function
    if main_function is not None:
        raise Exception("hyperdiv.run has already been called.")
    if not callable(main_fun):
        raise Exception("The argument to run() must be a function.")
    main_function = main_fun

    if not os.path.exists(static_directory):
        os.makedirs(static_directory)

    routes = [
        (r"/ws", ClientConnection),
        (
            r"/static/(.*)",
            HyperdivStaticFileHandler,
            {"path": static_directory},
        ),
        (
            r"/(.*)",
            HyperdivStaticFileHandler,
            {
                "path": get_public_folder_path(),
                "default_filename": "index.html",
            },
        ),
    ]

    if http_routes:
        for path, handler in http_routes.items():
            if path.startswith("/ws") or path.startswith("/static"):
                raise Exception(
                    'Custom server paths cannot start with "/ws" or "/static"'
                )
            routes.insert(0, (path, handler))

    app = Application(routes, debug=True)

    port = os.getenv("HD_PORT", 8888)

    try:
        port = int(port)
        if port < 1024:
            raise Exception("Port must be >= 1024")
    except Exception as e:
        logger.warn(f"Error parsing port: {str(e)}. Defaulting to 8888.")
        port = 8888

    def excepthook(exc_type, exc_value, exc_tb):
        traceback.print_exception(exc_type, exc_value, exc_tb)

    import sys

    sys.excepthook = excepthook

    def threading_excepthook(args):
        traceback.print_exception(args.exc_type, args.exc_value, args.exc_traceback)

    import threading

    threading.excepthook = threading_excepthook

    app.listen(port)

    profiling_enabled = os.getenv("HD_ENABLE_PROFILING", 0)
    if profiling_enabled:
        yappi.start()

    print(f"Open http://localhost:{port}/")
    try:
        ioloop.start()
    except KeyboardInterrupt:
        shutdown()

    if profiling_enabled:
        yappi.get_func_stats().print_all()
