#
# Utility methods for parsing rectangular props
#
def parse_rectangle_style(
    value,
    value_type_check,
    value_to_str,
    names=("top", "right", "bottom", "left"),
    default=None,
):
    """
    A helper method that parses a single value (e.g. a number),
    or a dict that has 1-4 values
    corresponding to the sides/corners of a rectangle.
    The default value is used when an entry is ommitted
    from the dictionary.
    """
    # First check if it's a valid single value
    if value_type_check(value):
        return value_to_str(value)

    # Otherwise check if it's a dict with valid entries
    elif (
        isinstance(value, dict)
        and len(value) > 0
        and len(value) <= 4
        and all(value_type_check(value[x]) for x in value)
    ):
        str_values = dict()
        for name in names:
            if name in value:
                str_values[name] = value_to_str(value[name])
            else:
                str_values[name] = value_to_str(default)
        return str_values

    # Finally check if it's a list/tuple with valid entries
    elif (
        isinstance(value, (list, tuple))
        and len(value) > 0
        and len(value) <= 4
        and all(value_type_check(x) for x in value)
    ):
        str_values = dict()
        for index, name in enumerate(names):
            if index >= len(value):
                str_values[name] = value_to_str(default)
            else:
                str_values[name] = value_to_str(value[index])
        return str_values
    raise ValueError("Invalid prop format.")


def is_non_negative_number(x):
    return type(x) in (int, float) and x >= 0


def is_percent(value):
    if type(value) is str and value.endswith("%"):
        num = float(value[:-1])
        return num >= 0 and num <= 100

    return False


def is_vwvh(value):
    if type(value) is str and (value.endswith("vw") or value.endswith("vh")):
        num = float(value[:-2])
        return num >= 0 and num <= 100

    return False


def is_fr(value):
    if type(value) is str and (value.endswith("fr")):
        num = float(value[:-2])
        return num >= 0 and num <= 100

    return False


def number_to_str(value, units="px"):
    return f"{value}{units}"


def is_hex_color(color):
    if type(color) is not str:
        return False
    if not color.startswith("#"):
        return False

    # Remove the leading "#"" symbol
    color = color[1:]

    # Check if it's rgb, rrggbb, rrggbbaa
    if len(color) not in (3, 6, 8):
        return False

    # Check if it's actually hex
    try:
        int(color, 16)
        return True
    except ValueError:
        return False


#
# Parse methods
#
def parse_padding(padding):
    """
    Accepts the following types:
    * 1. A number (int or float) that represnts rem units
    * 2. A dict that defines at least one of top, right, bottom, left
    * If an entry is omitted from the dict, it's default value is 0
    """
    try:
        return parse_rectangle_style(
            padding, is_non_negative_number, lambda x: x, default=0
        )
    except ValueError:
        raise ValueError("Invalid padding format.")


def parse_outline(outline):
    if outline is None or outline == "auto":
        return outline

    offset = 0
    if type(outline) in (list, tuple) and len(outline) == 4:
        offset = outline[3]
        outline = outline[:-1]

    return dict(outline=parse_border(outline), offset=number_to_str(offset))


def parse_border(border):
    """
    Accepts the following types:
    * 1. None
    * 2. Tuple with three values: (width in pixels , style , color)
    * 3. A dict that defines at least one of top, right, bottom, left
    *   where each entry is of type 1 or 2.
    * If an entry is omitted from the dict, it's default value is None.
    """
    try:
        return parse_rectangle_style(border, is_border, border_to_str, default=None)
    except ValueError:
        raise ValueError("Invalid border format.")


def parse_border_radius(radius):
    """
    Accepts the following types:
    * 1. A number (int or float)
    * 2. A dict that defines least one of topleft, topright, bottomleft, bottomright
    * If an entry is omitted from the dict, it's default value is 0.
    """
    try:
        return parse_rectangle_style(
            radius,
            is_non_negative_number,
            lambda x: x,
            names=("topleft", "topright", "bottomright", "bottomleft"),
            default=0,
        )
    except ValueError:
        raise ValueError("Invalid border radius format.")


def is_theme_color(color):
    if not isinstance(color, str):
        return False

    return theme_color_from_str(color) is not None


def theme_color_from_str(color):
    # Avoid cyclical imports
    from .theme import theme

    return getattr(theme(), color, None)


def is_color(color):
    return is_theme_color(color) or is_hex_color(color)


def parse_color(color):
    """
    Accepts the following types:
    * 1. None
    * 2. A hex color value #rrggbb, #rgb, or #rrggbbaa
    """
    if color is None:
        return None
    if is_hex_color(color):
        return color
    if is_theme_color(color):
        return theme_color_from_str(color)

    raise ValueError("Invalid color format.")


def parse_size(size):
    """
    Accepts the following types:
    * 1. None
    * 2. A number: int or float
    * 3. A percent string "x%", where 0 <= x <= 100
    """
    if size is None:
        return None
    if is_non_negative_number(size):
        return number_to_str(size)
    elif is_percent(size) or is_vwvh(size):
        return size
    elif size == "fit-content":
        return size
    raise ValueError("Invalid size format.")


def parse_width(width):
    return parse_size(width)


def parse_height(height):
    return parse_size(height)


def parse_cursor(cursor):
    if cursor in [None, "auto", "pointer"]:
        return cursor
    raise ValueError("Invalid cursor value.")


def parse_font_size(size):
    if size in [
        None,
        "xx-small",
        "x-small",
        "small",
        "medium",
        "large",
        "x-large",
        "xx-large",
        "xxx-large",
    ]:
        return size

    raise ValueError("Invalid font size.")


def parse_font_weight(weight):
    if weight in [None, "normal", "bold"]:
        return weight
    raise ValueError("Invalid font weight.")


def parse_font_family(values):
    if values is None:
        return None

    # Checks if the font is included
    if type(values) in (list, tuple):
        return ", ".join(values)

    return values


def parse_non_negative_number(num):
    if is_non_negative_number(num):
        return num
    else:
        raise ValueError("Prop must be a non negative number.")


def is_border(border):
    """
    Returns True if the border is None
    or a tuple/list with three entries (width, style, color)
    """
    return (
        border is None
        or (
            isinstance(border, (list, tuple))
            and len(border) == 3
            and is_non_negative_number(border[0])
            and is_border_style(border[1])
            and is_color(border[2])
        )
        or is_color(border)
        or is_non_negative_number(border)
        or is_border_style(border)
    )


def is_border_style(style):
    return style in ["solid", "dotted", "dashed"]


def border_to_str(border):
    from .theme import theme

    """
    Converts a tuple/list (width, style, color) into a single string
    that will be valid on the frontend.
    """
    if border is None:
        return None
    elif is_non_negative_number(border):
        thickness = border
        _, style, color = theme().input_border
        border = (thickness, style, color)
    elif is_color(border):
        color = parse_color(border)
        thickness, style, _ = theme().input_border
        border = (thickness, color, style)
    elif is_border_style(border):
        style = border
        thickness, _, color = theme().input_border
        border = (thickness, style, color)

    return f"{number_to_str(border[0])} {border[1]} {border[2]}"


def to_rgba(color):
    if is_theme_color(color):
        color = theme_color_from_str(color)

    color = color[1:]
    r, g, b, a = ("0", "0", "0", "FF")
    if len(color) == 3:
        r = color[0] * 2
        g = color[1] * 2
        b = color[2] * 2
    elif len(color) >= 6:
        r = color[0:2]
        g = color[2:4]
        b = color[4:6]
    if len(color) == 8:
        a = color[6:8]

    return (int(r, 16), int(g, 16), int(b, 16), int(a, 16))


def is_light(rgba):
    r, b, g, a = rgba
    return (r + g + b) / 3 >= 128


def offset_color(rgba, offset):
    if rgba is None:
        return None
    else:
        r, g, b, a = rgba
        r = min(255, max(0, r + offset))
        g = min(255, max(0, g + offset))
        b = min(255, max(0, b + offset))

        return rgba_to_hex((r, g, b, a))


def rgba_to_hex(rgba):
    r, g, b, a = rgba
    return "#%02x%02x%02x%02x" % (r, g, b, a)


def highlight(color, amount=50):
    rgba = to_rgba(color)
    if is_light(rgba):
        return darken(rgba, amount=amount)
    else:
        return lighten(rgba, amount=amount)


def lighten(color, amount=50):
    if type(color) is str:
        color = to_rgba(color)
    return offset_color(color, amount)


def darken(color, amount=50):
    if type(color) is str:
        color = to_rgba(color)
    return offset_color(color, -amount)
