from .style_utils import (
    parse_border,
    parse_border_radius,
    parse_color,
    parse_font_family,
    parse_padding,
)
from ..core.element import Element, Prop


def parse_mode(mode):
    if mode in ("system", "dark", "light"):
        return mode
    raise ValueError("Invalid theme mode.")


def light_or_dark(propname):
    def getter(theme_element):
        mode = "dark" if theme_element.is_dark else "light"
        return getattr(theme_element, f"{mode}_{propname}")

    return getter


class theme(Element):
    is_singleton = True
    is_collected = False

    # Dark / light mode settings
    system_mode = Prop("dark")  # TODO: only assignable from the frontend.
    mode = Prop("system", parse=parse_mode)

    # These remain the same in both light and dark mode.
    blue = Prop("#4A5AD3", parse=parse_color)
    green = Prop("#00999C", parse=parse_color)
    yellow = Prop("#F4BD59", parse=parse_color)
    red = Prop("#EE6869", parse=parse_color)
    black = Prop("#000000", parse=parse_color)
    white = Prop("#ffffff", parse=parse_color)

    # Aliases for the above
    primary_color = blue
    error_color = red
    warning_color = yellow
    success_color = green

    font_family = Prop("Nunito", parse=parse_font_family)
    border_radius = Prop(0, parse=parse_border_radius)
    input_padding = Prop(8, parse=parse_padding)

    # Dark mode values
    dark_background_color = Prop("#030405", parse=parse_color)
    dark_surface_color = Prop("#2C2C2D", parse=parse_color)
    dark_font_color = Prop("#fff", parse=parse_color)
    dark_container_border = Prop((1, "solid", "#545555"), parse=parse_border)
    dark_input_border = Prop((1, "solid", "#545555"), parse=parse_border)
    dark_modal_cover_color = Prop("#ffffff18", parse=parse_color)

    # Light mode values
    light_background_color = Prop("#fff", parse=parse_color)
    light_surface_color = Prop("#FAFAFA", parse=parse_color)
    light_font_color = Prop("#030405", parse=parse_color)
    light_container_border = Prop((1, "solid", "#CFD0D1"), parse=parse_border)
    light_input_border = Prop((1, "solid", "#CFD0D1"), parse=parse_border)
    light_modal_cover_color = Prop("#00000018", parse=parse_color)

    # Computed props based on the mode
    # TODO: These should be read-only
    background_color = Prop(light_or_dark("background_color"), parse=parse_color)
    surface_color = Prop(light_or_dark("surface_color"), parse=parse_color)
    font_color = Prop(light_or_dark("font_color"), parse=parse_color)
    container_border = Prop(light_or_dark("container_border"), parse=parse_border)
    input_border = Prop(light_or_dark("input_border"), parse=parse_border)
    modal_cover_color = Prop(light_or_dark("modal_cover_color"), parse=parse_color)

    @property
    def is_dark(self):
        return self.mode == "dark" or (
            self.mode == "system" and self.system_mode == "dark"
        )


def theme_prop(propname):
    existing_prop = getattr(theme, propname)
    return Prop(lambda e: getattr(theme(), propname), parse=existing_prop.parse)
